use sprinklers_core::model::{ZoneId, ZoneRef};
use sprinklers_core::zone_interface::ZoneInterface;

use actix::{
    Actor, ActorContext, Addr, AsyncContext, Handler, Message, MessageResult, SpawnHandle,
};
use futures_util::TryFutureExt;
use std::{
    future::Future,
    mem::swap,
    sync::{
        atomic::{AtomicI32, Ordering},
        Arc,
    },
    time::Duration,
};
use thiserror::Error;
use tokio::{
    sync::{broadcast, watch},
    time::Instant,
};
use tracing::{debug, trace, warn};

#[derive(Debug, Clone, PartialEq, Eq, Hash, serde::Deserialize, serde::Serialize)]
pub struct ZoneRunHandle(i32);

impl ZoneRunHandle {
    pub fn into_inner(self) -> i32 {
        self.0
    }
}

#[derive(Clone, Debug)]
pub enum ZoneEvent {
    RunStart(ZoneRunHandle, ZoneRef),
    RunFinish(ZoneRunHandle, ZoneRef),
    RunPause(ZoneRunHandle, ZoneRef),
    RunUnpause(ZoneRunHandle, ZoneRef),
    RunCancel(ZoneRunHandle, ZoneRef),
    RunnerPause,
    RunnerUnpause,
}

pub type ZoneEventRecv = broadcast::Receiver<ZoneEvent>;
type ZoneEventSend = broadcast::Sender<ZoneEvent>;

const EVENT_CAPACITY: usize = 8;

#[derive(Clone, Debug, PartialEq)]
pub enum ZoneRunState {
    Waiting,
    Running {
        start_time: Instant,
    },
    Finished,
    Cancelled,
    Paused {
        start_time: Instant,
        pause_time: Instant,
    },
}

#[derive(Clone, Debug)]
pub struct ZoneRun {
    pub handle: ZoneRunHandle,
    pub zone: ZoneRef,
    pub duration: Duration,
    pub total_duration: Duration,
    pub state: ZoneRunState,
}

impl ZoneRun {
    fn new(handle: ZoneRunHandle, zone: ZoneRef, duration: Duration) -> Self {
        Self {
            handle,
            zone,
            duration,
            total_duration: duration,
            state: ZoneRunState::Waiting,
        }
    }

    pub fn is_running(&self) -> bool {
        matches!(self.state, ZoneRunState::Running{..})
    }

    #[allow(dead_code)]
    pub fn is_paused(&self) -> bool {
        matches!(self.state, ZoneRunState::Paused{..})
    }
}

pub type ZoneRunQueue = im::Vector<Arc<ZoneRun>>;

#[derive(Clone, Debug)]
pub struct ZoneRunnerState {
    pub run_queue: ZoneRunQueue,
    pub paused: bool,
}

impl Default for ZoneRunnerState {
    fn default() -> Self {
        Self {
            run_queue: ZoneRunQueue::default(),
            paused: false,
        }
    }
}

pub type ZoneRunnerStateRecv = watch::Receiver<ZoneRunnerState>;

struct ZoneRunnerInner {
    interface: Arc<dyn ZoneInterface>,
    event_send: Option<ZoneEventSend>,
    state_send: watch::Sender<ZoneRunnerState>,
    delay_future: Option<SpawnHandle>,
    did_change: bool,
}

impl ZoneRunnerInner {
    fn send_event(&mut self, event: ZoneEvent) {
        if let Some(event_send) = &mut self.event_send {
            match event_send.send(event) {
                Ok(_) => {}
                Err(_closed) => {
                    self.event_send = None;
                }
            }
        }
    }

    fn subscribe_event(&mut self) -> ZoneEventRecv {
        match &mut self.event_send {
            Some(event_send) => event_send.subscribe(),
            None => {
                let (event_send, event_recv) = broadcast::channel(EVENT_CAPACITY);
                self.event_send = Some(event_send);
                event_recv
            }
        }
    }

    fn start_run(&mut self, run: &mut Arc<ZoneRun>) {
        use ZoneRunState::*;
        let run = Arc::make_mut(run);
        debug!(zone_id = run.zone.id, "starting running zone");
        self.interface.set_zone_state(run.zone.interface_id, true);
        run.state = Running {
            start_time: Instant::now(),
        };
        self.send_event(ZoneEvent::RunStart(run.handle.clone(), run.zone.clone()));
        self.did_change = true;
    }

    fn finish_run(&mut self, run: &mut Arc<ZoneRun>) {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(zone_id = run.zone.id, "finished running zone");
            self.interface.set_zone_state(run.zone.interface_id, false);
            run.state = ZoneRunState::Finished;
            self.send_event(ZoneEvent::RunFinish(run.handle.clone(), run.zone.clone()));
            self.did_change = true;
        } else {
            warn!(
                zone_id = run.zone.id,
                state = debug(&run.state),
                "cannot finish run which is not running"
            );
        }
    }

    fn cancel_run(&mut self, run: &mut Arc<ZoneRun>) -> bool {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(zone_id = run.zone.id, "cancelling running zone");
            self.interface.set_zone_state(run.zone.interface_id, false);
        }
        if run.state != ZoneRunState::Cancelled {
            run.state = ZoneRunState::Cancelled;
            self.send_event(ZoneEvent::RunCancel(run.handle.clone(), run.zone.clone()));
            self.did_change = true;
            true
        } else {
            false
        }
    }

    fn pause_run(&mut self, run: &mut Arc<ZoneRun>) {
        use ZoneRunState::*;
        let run = Arc::make_mut(run);
        let new_state = match run.state {
            Running { start_time } => {
                debug!(zone_id = run.zone.id, "pausing running zone");
                self.interface.set_zone_state(run.zone.interface_id, false);
                Paused {
                    start_time,
                    pause_time: Instant::now(),
                }
            }
            Waiting => {
                debug!(zone_id = run.zone.id, "pausing waiting zone");
                Paused {
                    start_time: Instant::now(),
                    pause_time: Instant::now(),
                }
            }
            Finished | Cancelled | Paused { .. } => {
                return;
            }
        };
        run.state = new_state;
        self.send_event(ZoneEvent::RunPause(run.handle.clone(), run.zone.clone()));
        self.did_change = true;
    }

    fn unpause_run(&mut self, run: &mut Arc<ZoneRun>) {
        use ZoneRunState::*;
        let run = Arc::make_mut(run);
        match run.state {
            Paused {
                start_time,
                pause_time,
            } => {
                debug!(zone_id = run.zone.id, "unpausing zone");
                self.interface.set_zone_state(run.zone.interface_id, true);
                run.state = Running {
                    start_time: Instant::now(),
                };
                let ran_for = pause_time - start_time;
                run.duration -= ran_for;
                self.send_event(ZoneEvent::RunUnpause(run.handle.clone(), run.zone.clone()));
            }
            Waiting | Finished | Cancelled | Running { .. } => {
                warn!(
                    zone_id = run.zone.id,
                    state = debug(&run.state),
                    "can only unpause paused zone"
                );
            }
        }
        self.did_change = true;
    }

    fn process_after_delay(
        &mut self,
        after: Duration,
        ctx: &mut <ZoneRunnerActor as Actor>::Context,
    ) {
        let delay_future = ctx.notify_later(Process, after);
        if let Some(old_future) = self.delay_future.replace(delay_future) {
            ctx.cancel_future(old_future);
        }
    }

    fn cancel_process(&mut self, ctx: &mut <ZoneRunnerActor as Actor>::Context) {
        if let Some(old_future) = self.delay_future.take() {
            ctx.cancel_future(old_future);
        }
    }
}

struct ZoneRunnerActor {
    state: ZoneRunnerState,
    inner: ZoneRunnerInner,
}

impl Actor for ZoneRunnerActor {
    type Context = actix::Context<Self>;

    fn started(&mut self, _ctx: &mut Self::Context) {
        trace!("zone_runner starting");
        for i in 0..self.inner.interface.num_zones() {
            self.inner.interface.set_zone_state(i, false);
        }
    }

    fn stopped(&mut self, _ctx: &mut Self::Context) {
        trace!("zone_runner stopped");
        for i in 0..self.inner.interface.num_zones() {
            self.inner.interface.set_zone_state(i, false);
        }
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct Quit;

impl Handler<Quit> for ZoneRunnerActor {
    type Result = ();

    fn handle(&mut self, _msg: Quit, ctx: &mut Self::Context) -> Self::Result {
        ctx.stop();
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct QueueRun(ZoneRunHandle, ZoneRef, Duration);

impl Handler<QueueRun> for ZoneRunnerActor {
    type Result = ();

    fn handle(&mut self, msg: QueueRun, ctx: &mut Self::Context) -> Self::Result {
        let QueueRun(handle, zone, duration) = msg;

        let run: Arc<ZoneRun> = ZoneRun::new(handle, zone, duration).into();
        self.state.run_queue.push_back(run);
        self.inner.did_change = true;

        ctx.notify(Process);
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "bool")]
struct CancelRun(ZoneRunHandle);

impl Handler<CancelRun> for ZoneRunnerActor {
    type Result = bool;

    fn handle(&mut self, msg: CancelRun, ctx: &mut Self::Context) -> Self::Result {
        let CancelRun(handle) = msg;
        let mut cancelled = false;
        for run in self
            .state
            .run_queue
            .iter_mut()
            .filter(|run| run.handle == handle)
        {
            trace!(handle = handle.0, "cancelling run by handle");
            cancelled = self.inner.cancel_run(run);
        }
        ctx.notify(Process);
        cancelled
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "usize")]
struct CancelByZone(ZoneId);

impl Handler<CancelByZone> for ZoneRunnerActor {
    type Result = usize;

    fn handle(&mut self, msg: CancelByZone, ctx: &mut Self::Context) -> Self::Result {
        let CancelByZone(zone_id) = msg;
        let mut count = 0_usize;
        for run in self
            .state
            .run_queue
            .iter_mut()
            .filter(|run| run.zone.id == zone_id)
        {
            trace!(handle = run.handle.0, zone_id, "cancelling run by zone");
            if self.inner.cancel_run(run) {
                count += 1;
            }
        }
        ctx.notify(Process);
        count
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "usize")]
struct CancelAll;

impl Handler<CancelAll> for ZoneRunnerActor {
    type Result = usize;

    fn handle(&mut self, _msg: CancelAll, ctx: &mut Self::Context) -> Self::Result {
        let mut old_runs = ZoneRunQueue::new();
        swap(&mut old_runs, &mut self.state.run_queue);
        trace!(count = old_runs.len(), "cancelling all runs");
        let mut count = 0usize;
        for mut run in old_runs {
            if self.inner.cancel_run(&mut run) {
                count += 1;
            }
        }
        ctx.notify(Process);
        count
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct SetPaused(bool);

impl Handler<SetPaused> for ZoneRunnerActor {
    type Result = ();

    fn handle(&mut self, msg: SetPaused, ctx: &mut Self::Context) -> Self::Result {
        let SetPaused(pause) = msg;
        if pause != self.state.paused {
            if pause {
                self.state.paused = true;
                self.inner.send_event(ZoneEvent::RunnerPause);
            } else {
                self.state.paused = false;
                self.inner.send_event(ZoneEvent::RunnerUnpause);
            }
            self.inner.did_change = true;
            ctx.notify(Process);
        }
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "ZoneEventRecv")]
struct Subscribe;

impl Handler<Subscribe> for ZoneRunnerActor {
    type Result = MessageResult<Subscribe>;

    fn handle(&mut self, _msg: Subscribe, _ctx: &mut Self::Context) -> Self::Result {
        let event_recv = self.inner.subscribe_event();
        MessageResult(event_recv)
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct Process;

impl Handler<Process> for ZoneRunnerActor {
    type Result = ();

    fn handle(&mut self, _msg: Process, ctx: &mut Self::Context) -> Self::Result {
        self.process(ctx)
    }
}

impl ZoneRunnerActor {
    fn new(interface: Arc<dyn ZoneInterface>, state_send: watch::Sender<ZoneRunnerState>) -> Self {
        Self {
            state: ZoneRunnerState::default(),
            inner: ZoneRunnerInner {
                interface,
                event_send: None,
                state_send,
                delay_future: None,
                did_change: false,
            },
        }
    }

    fn process_queue(&mut self, ctx: &mut actix::Context<Self>) {
        use ZoneRunState::*;
        let state = &mut self.state;
        while let Some(current_run) = state.run_queue.front_mut() {
            let run_finished = match (&current_run.state, state.paused) {
                (Waiting, false) => {
                    self.inner.start_run(current_run);
                    self.inner.process_after_delay(current_run.duration, ctx);
                    false
                }
                (Running { start_time }, false) => {
                    let time_to_finish = start_time.elapsed() >= current_run.duration;
                    if time_to_finish {
                        self.inner.finish_run(current_run);
                        self.inner.cancel_process(ctx);
                    }
                    time_to_finish
                }
                (Paused { .. }, false) => {
                    self.inner.unpause_run(current_run);
                    self.inner.process_after_delay(current_run.duration, ctx);
                    false
                }
                (Waiting, true) => {
                    self.inner.pause_run(current_run);
                    false
                }
                (Running { .. }, true) => {
                    self.inner.pause_run(current_run);
                    self.inner.cancel_process(ctx);
                    false
                }
                (Paused { .. }, true) => false,
                (Cancelled, _) | (Finished, _) => true,
            };

            if run_finished {
                state.run_queue.pop_front();
            } else {
                break;
            }
        }
    }

    fn process(&mut self, ctx: &mut actix::Context<Self>) {
        self.process_queue(ctx);

        if self.inner.did_change {
            let _ = self.inner.state_send.broadcast(self.state.clone());
            self.inner.did_change = false;
        }
    }
}

#[derive(Debug, Clone, Error)]
#[error("error communicating with ZoneRunner: {0}")]
pub struct Error(#[from] actix::MailboxError);

pub type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(Clone)]
pub struct ZoneRunner {
    state_recv: ZoneRunnerStateRecv,
    addr: Addr<ZoneRunnerActor>,
    next_run_id: Arc<AtomicI32>,
}

#[allow(dead_code)]
impl ZoneRunner {
    pub fn new(interface: Arc<dyn ZoneInterface>) -> Self {
        let (state_send, state_recv) = watch::channel(ZoneRunnerState::default());
        let addr = ZoneRunnerActor::new(interface, state_send).start();
        Self {
            state_recv,
            addr,
            next_run_id: Arc::new(AtomicI32::new(1)),
        }
    }

    pub fn quit(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(Quit).map_err(From::from)
    }

    fn queue_run_inner(&mut self, zone: ZoneRef, duration: Duration) -> (QueueRun, ZoneRunHandle) {
        let run_id = self.next_run_id.fetch_add(1, Ordering::SeqCst);
        let handle = ZoneRunHandle(run_id);
        (QueueRun(handle.clone(), zone, duration), handle)
    }

    pub fn do_queue_run(&mut self, zone: ZoneRef, duration: Duration) -> ZoneRunHandle {
        let (queue_run, handle) = self.queue_run_inner(zone, duration);
        self.addr.do_send(queue_run);
        handle
    }

    pub fn queue_run(
        &mut self,
        zone: ZoneRef,
        duration: Duration,
    ) -> impl Future<Output = Result<ZoneRunHandle>> {
        let (queue_run, handle) = self.queue_run_inner(zone, duration);
        self.addr
            .send(queue_run)
            .map_err(From::from)
            .map_ok(move |_| handle)
    }

    pub fn do_cancel_run(&mut self, handle: ZoneRunHandle) {
        self.addr.do_send(CancelRun(handle))
    }

    pub fn cancel_run(&mut self, handle: ZoneRunHandle) -> impl Future<Output = Result<bool>> {
        self.addr.send(CancelRun(handle)).map_err(From::from)
    }

    pub fn cancel_by_zone(&mut self, zone_id: ZoneId) -> impl Future<Output = Result<usize>> {
        self.addr.send(CancelByZone(zone_id)).map_err(From::from)
    }

    pub fn cancel_all(&mut self) -> impl Future<Output = Result<usize>> {
        self.addr.send(CancelAll).map_err(From::from)
    }

    pub fn pause(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(SetPaused(true)).map_err(From::from)
    }

    pub fn unpause(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(SetPaused(false)).map_err(From::from)
    }

    pub fn subscribe(&mut self) -> impl Future<Output = Result<ZoneEventRecv>> {
        self.addr.send(Subscribe).map_err(From::from)
    }

    pub fn get_state_recv(&self) -> ZoneRunnerStateRecv {
        self.state_recv.clone()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::trace_listeners::{EventListener, Filters};
    use sprinklers_core::{
        model::{Zone, Zones},
        zone_interface::MockZoneInterface,
    };

    use assert_matches::assert_matches;
    use im::ordmap;
    use tracing_subscriber::prelude::*;

    #[actix_rt::test]
    async fn test_quit() {
        let quit_msg = EventListener::new(
            Filters::new()
                .target("sprinklers_actors::zone_runner")
                .message("zone_runner stopped"),
        );
        let subscriber = tracing_subscriber::registry().with(quit_msg.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        let interface = MockZoneInterface::new(6);
        let mut runner = ZoneRunner::new(Arc::new(interface));
        tokio::task::yield_now().await;
        runner.quit().await.unwrap();

        assert_eq!(quit_msg.get_count(), 1);
    }

    fn make_zones_and_interface() -> (Zones, Arc<MockZoneInterface>) {
        let interface = Arc::new(MockZoneInterface::new(2));
        let zones: Zones = ordmap![
            1 => Zone {
                id: 1,
                name: "Zone 1".into(),
                interface_id: 0,
            }.into(),
            2 => Zone {
                id: 2,
                name: "Zone 2".into(),
                interface_id: 1,
            }.into()
        ];
        (zones, interface)
    }

    fn assert_zone_states(interface: &MockZoneInterface, states: &[bool]) {
        for (id, state) in states.iter().enumerate() {
            assert_eq!(
                interface.get_zone_state(id as u32),
                *state,
                "zone interface id {} did not match",
                id
            );
        }
    }

    async fn advance(dur: Duration) {
        // HACK: advance should really be enough, but we need another yield_now
        tokio::time::pause();
        tokio::time::advance(dur).await;
        tokio::task::yield_now().await;
        tokio::time::resume();
    }

    #[actix_rt::test]
    async fn test_queue() {
        let (zones, interface) = make_zones_and_interface();
        let mut runner = ZoneRunner::new(interface.clone());

        assert_zone_states(&interface, &[false, false]);

        // Queue single zone, make sure it runs
        runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_zone_states(&interface, &[true, false]);

        advance(Duration::from_secs(11)).await;

        assert_zone_states(&interface, &[false, false]);

        // Queue two zones, make sure they run one at a time
        runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_zone_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;

        assert_zone_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_zone_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_cancel_run() {
        let (zones, interface) = make_zones_and_interface();
        let mut runner = ZoneRunner::new(interface.clone());

        let run1 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run2 = runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_zone_states(&interface, &[false, true]);

        runner.cancel_run(run1).await.unwrap();
        tokio::task::yield_now().await;

        assert_zone_states(&interface, &[true, false]);

        runner.cancel_run(run3).await.unwrap();
        advance(Duration::from_secs(11)).await;

        assert_zone_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_cancel_all() {
        let (zones, interface) = make_zones_and_interface();
        let mut runner = ZoneRunner::new(interface.clone());

        runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, true]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, false]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_pause() {
        let (zones, interface) = make_zones_and_interface();
        let mut runner = ZoneRunner::new(interface.clone());

        let _run1 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run3 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, true]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, false]);

        advance(Duration::from_secs(10)).await;
        assert_zone_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, true]);

        advance(Duration::from_secs(8)).await;
        assert_zone_states(&interface, &[false, true]);

        advance(Duration::from_secs(2)).await;
        assert_zone_states(&interface, &[true, false]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, false]);

        // cancel paused run
        runner.cancel_run(run2).await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_zone_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;
        assert_zone_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_event() {
        let (zones, interface) = make_zones_and_interface();
        let mut runner = ZoneRunner::new(interface.clone());

        let mut event_recv = runner.subscribe().await.unwrap();

        let run1 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(zones[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(zones[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        assert_matches!(
            event_recv.recv().await,
            Ok(ZoneEvent::RunStart(handle, _))
            if handle == run1
        );

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunPause(handle, _)) if handle == run1);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunUnpause(handle, _)) if handle == run1);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunFinish(handle, _)) if handle == run1);
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunStart(handle, _)) if handle == run2);

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunPause(handle, _)) if handle == run2);

        // cancel paused run
        runner.cancel_run(run2.clone()).await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunCancel(handle, _)) if handle == run2);
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunPause(handle, _)) if handle == run3);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunUnpause(handle, _)) if handle == run3);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(ZoneEvent::RunFinish(handle, _)) if handle == run3);

        runner.quit().await.unwrap();
    }
}
