use std::iter::repeat_with;
use std::sync::atomic::{AtomicBool, Ordering};
use tracing::debug;

pub type ZoneNum = u32;

pub trait ZoneInterface: Send + Sync {
    fn num_zones(&self) -> ZoneNum;
    fn set_zone_state(&self, id: ZoneNum, running: bool);
    fn get_zone_state(&self, id: ZoneNum) -> bool;
}

pub struct MockZoneInterface {
    states: Vec<AtomicBool>,
}

impl MockZoneInterface {
    #[allow(dead_code)]
    pub fn new(num_zones: ZoneNum) -> Self {
        Self {
            states: repeat_with(|| AtomicBool::new(false))
                .take(num_zones as usize)
                .collect(),
        }
    }
}

impl ZoneInterface for MockZoneInterface {
    fn num_zones(&self) -> ZoneNum {
        self.states.len() as ZoneNum
    }
    fn set_zone_state(&self, id: ZoneNum, running: bool) {
        debug!(id, running, "setting zone");
        self.states[id as usize].store(running, Ordering::SeqCst);
    }
    fn get_zone_state(&self, id: ZoneNum) -> bool {
        self.states[id as usize].load(Ordering::SeqCst)
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_mock_zone_interface() {
        let iface = MockZoneInterface::new(6);
        assert_eq!(iface.num_zones(), 6);
        for i in 0..6u32 {
            assert_eq!(iface.get_zone_state(i), false);
        }
        for i in 0..6u32 {
            iface.set_zone_state(i, true);
        }
        for i in 0..6u32 {
            assert_eq!(iface.get_zone_state(i), true);
        }
        for i in 0..6u32 {
            iface.set_zone_state(i, false);
        }
        for i in 0..6u32 {
            assert_eq!(iface.get_zone_state(i), false);
        }
    }
}
