mod actor;
mod event_loop;
mod request;
mod topics;
mod update_listener;
mod zone_runner_json;

pub use request::RequestContext;
pub use update_listener::UpdateListener;

use self::topics::{CollectionTopics, Topics};
use sprinklers_actors::zone_runner::ZoneRunnerState;
use sprinklers_core::model::{ProgramId, Programs, ZoneId, Zones};
use zone_runner_json::ZoneRunnerStateJson;

use actix::{Actor, Addr};
use eyre::WrapErr;
use rumqttc::{LastWill, MqttOptions, QoS};
use std::{
    marker::PhantomData,
    ops::{Deref, DerefMut},
    sync::Arc,
};

#[derive(Clone, Debug)]
pub struct Options {
    pub broker_host: String,
    pub broker_port: u16,
    pub device_id: String,
    pub client_id: String,
}

#[derive(Clone)]
pub struct MqttInterface {
    client: rumqttc::AsyncClient,
    topics: Topics<Arc<str>>,
}

impl MqttInterface {
    fn new(options: Options) -> (Self, rumqttc::EventLoop) {
        let mqtt_prefix = format!("devices/{}", options.device_id);
        let topics: Topics<Arc<str>> = Topics::new(mqtt_prefix.into());
        let mut mqtt_opts =
            MqttOptions::new(options.client_id, options.broker_host, options.broker_port);

        let last_will = LastWill::new(topics.connected(), "false", QoS::AtLeastOnce, true);
        mqtt_opts.set_last_will(last_will);

        let (client, event_loop) = rumqttc::AsyncClient::new(mqtt_opts, 16);

        (Self { client, topics }, event_loop)
    }

    async fn publish_data<P>(&mut self, topic: String, payload: &P) -> eyre::Result<()>
    where
        P: serde::Serialize,
    {
        let payload_vec =
            serde_json::to_vec(payload).wrap_err("failed to serialize publish payload")?;
        self.client
            .publish(topic, QoS::AtLeastOnce, true, payload_vec)
            .await
            .wrap_err("failed to publish")?;
        Ok(())
    }

    async fn publish_connected(&mut self, connected: bool) -> eyre::Result<()> {
        self.publish_data(self.topics.connected(), &connected)
            .await
            .wrap_err("failed to publish connected topic")
    }

    async fn cancel(&mut self) -> Result<(), rumqttc::ClientError> {
        self.client.cancel().await
    }

    pub fn zones(&mut self) -> MqttCollection<'_, topics::ZoneTopics, Zones> {
        MqttCollection::new(self)
    }

    pub fn programs(&mut self) -> MqttCollection<'_, topics::ProgramTopics, Programs> {
        MqttCollection::new(self)
    }

    pub async fn publish_zone_runner(&mut self, sr_state: &ZoneRunnerState) -> eyre::Result<()> {
        let json: ZoneRunnerStateJson = sr_state.into();
        self.publish_data(self.topics.zone_runner(), &json)
            .await
            .wrap_err("failed to publish zone runner")
    }

    async fn publish_response(&mut self, resp: request::ResponseWithId) -> eyre::Result<()> {
        let payload_vec =
            serde_json::to_vec(&resp).wrap_err("failed to serialize request response")?;
        // TODO: if couldn't serialize, just in case can have a static response
        self.client
            .publish(self.topics.responses(), QoS::AtMostOnce, false, payload_vec)
            .await
            .wrap_err("failed to publish request response")?;
        Ok(())
    }

    async fn subscribe_requests(&mut self) -> eyre::Result<()> {
        self.client
            .subscribe(self.topics.requests(), QoS::ExactlyOnce)
            .await?;
        Ok(())
    }
}

pub struct MqttCollection<'a, T, U> {
    client: &'a mut rumqttc::AsyncClient,
    topics: T,
    collection: PhantomData<U>,
}

impl<'a, T: CollectionTopics<'a>, U> MqttCollection<'a, T, U> {
    fn new(interface: &'a mut MqttInterface) -> Self {
        Self {
            client: &mut interface.client,
            topics: T::new(interface.topics.prefix()),
            collection: PhantomData,
        }
    }

    async fn publish<P: serde::Serialize>(
        &mut self,
        topic: String,
        payload: &P,
    ) -> eyre::Result<()> {
        let payload_vec =
            serde_json::to_vec(payload).wrap_err("failed to serialize publish payload")?;
        self.client
            .publish(topic, QoS::AtLeastOnce, true, payload_vec)
            .await
            .wrap_err("failed to publish")?;
        Ok(())
    }

    async fn publish_ids_impl(&mut self, ids: &[u32]) -> eyre::Result<()> {
        self.publish(self.topics.ids(), &ids).await?;
        Ok(())
    }

    async fn publish_data<V: serde::Serialize>(&mut self, id: u32, item: &V) -> eyre::Result<()> {
        self.publish(self.topics.data(id), item).await?;
        Ok(())
    }
}

impl<'a, T: CollectionTopics<'a>, V: serde::Serialize>
    MqttCollection<'a, T, im::OrdMap<u32, Arc<V>>>
{
    async fn publish_ids(&mut self, items: &im::OrdMap<u32, Arc<V>>) -> eyre::Result<()> {
        let ids: Vec<u32> = items.keys().cloned().collect();
        self.publish_ids_impl(&ids).await
    }

    pub async fn publish_diff(
        &mut self,
        old_values: Option<&im::OrdMap<u32, Arc<V>>>,
        new_values: &im::OrdMap<u32, Arc<V>>,
    ) -> eyre::Result<()> {
        let mut published_ids = false;
        for (id, value) in new_values {
            let new_value_different = old_values
                .and_then(|old_values| old_values.get(id))
                .map(|old_value| !Arc::ptr_eq(old_value, value));
            let publish_value = if let Some(different) = new_value_different {
                different
            } else {
                // old value does not exist
                if !published_ids {
                    self.publish_ids(new_values).await?;
                    published_ids = true;
                }
                true
            };
            if publish_value {
                self.publish_data(*id, &**value).await?;
            }
        }
        Ok(())
    }

    pub async fn publish_all(&mut self, values: &im::OrdMap<u32, Arc<V>>) -> eyre::Result<()> {
        self.publish_diff(None, values).await
    }
}

impl<'a> MqttCollection<'a, topics::ZoneTopics<'a>, Zones> {
    // Zone state can be derived from zone runner state...
    pub async fn publish_state(&mut self, zone_id: ZoneId, state: bool) -> eyre::Result<()> {
        self.publish(self.topics.state(zone_id), &state)
            .await
            .wrap_err("failed to publish zone state")
    }
}

impl<'a> MqttCollection<'a, topics::ProgramTopics<'a>, Programs> {
    pub async fn publish_running(
        &mut self,
        program_id: ProgramId,
        running: bool,
    ) -> eyre::Result<()> {
        self.publish(self.topics.running(program_id), &running)
            .await
            .wrap_err("failed to publish program running")
    }

    pub async fn publish_next_run(
        &mut self,
        program_id: ProgramId,
        next_run: chrono::DateTime<chrono::Local>,
    ) -> eyre::Result<()> {
        let payload = next_run.to_rfc3339();
        self.publish(self.topics.next_run(program_id), &payload)
            .await
            .wrap_err("failed to publish program next run")
    }
}

pub struct MqttInterfaceTask {
    interface: MqttInterface,
    addr: Addr<actor::MqttActor>,
}

impl MqttInterfaceTask {
    pub fn start(options: Options, request_context: RequestContext) -> Self {
        let (interface, event_loop) = MqttInterface::new(options);

        let addr = actor::MqttActor::new(interface.clone(), event_loop, request_context).start();

        Self { interface, addr }
    }

    pub async fn quit(self) -> eyre::Result<()> {
        self.addr.send(actor::Quit).await?;
        Ok(())
    }
}

impl Deref for MqttInterfaceTask {
    type Target = MqttInterface;

    fn deref(&self) -> &Self::Target {
        &self.interface
    }
}

impl DerefMut for MqttInterfaceTask {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.interface
    }
}
